# Copyright(c) 2002-2006 Association of Universities for Research in Astronomy, Inc.

procedure gdisplay (image, frame)

# Display GMOS (raw) images
#
# Version Feb 28, 2002   IJ  v1.3 release
#         Apr 26, 2002   IJ  change warning comment, handle [1] images
#         Jul  7, 2002   IJ  handle dir-names+short image names correctly
#         Sept 20, 2002  IJ   v1.4 release
#         Oct 17, 2002   BM  generalized for GMOS-S, 6amps, and all readmodes
#         Oct 22, 2002   BM  fix problem with defining l_gainmode when fl_bias-
#         Jan 31, 2003   BM  Added output file
#         Sep 13, 2006   JH  Add image number feature as input image if current UT
#                            
char    image       {"",prompt="GMOS image to display, can use number if current UT"} # OLDP-1-input-primary-none
int     frame       {1,prompt="Frame to write to"}                                 # OLDP-1
char    output      {"",prompt="Save pasted file to this name if not blank"}       # OLDP-3
bool    fl_paste    {no,prompt="Paste images to one for imexamine"}                # OLDP-2
bool    fl_bias     {no,prompt="Rough bias subtraction"}                           # OLDP-2
char    rawpath     {"adata$",prompt="Path for input image if not included in name"}
int     gap         {37,prompt="Size of the gap between the CCDs (in pixels)"}     # OLDP-3
real    z1          {0.,prompt="Lower limit if not autoscaling"}                   # OLDP-2
real    z2          {0.,prompt="Upper limit if not autoscaling"}                   # OLDP-2
bool    fl_sat      {no,prompt="Flag saturated pixels"}                            # OLDP-2
bool    fl_imexam   {yes,prompt="If possible, run imexam"}
real    signal      {INDEF,prompt="Flag pixels with signal above this limit"}      # OLDP-2
char    sci_ext     {"all",prompt="Name of extension(s) to display"}               # OLDP-3
char    observatory {")_.observatory", prompt="Observatory (gemini-north or gemini-south)"}
char    prefix      {"auto",prompt="File prefix, (N/S)YYYYMMDDS if not auto"}
char    key_detsec  {"DETSEC",prompt="Header keyword for detector section"}        # OLDP-3
char    key_datasec {"DATASEC",prompt="Header keyword for data section"}           # OLDP-3
char    key_ccdsum  {"CCDSUM",prompt="Header keyword for CCD binning"}             # OLDP-3
char    gaindb      {"default",prompt="Database with gain data"}    # OLDP-2
bool    verbose     {no,prompt="Verbose"}                                          # OLDP-4
int     status      {0,prompt="Exit status (0=good)"}
struct  *scanfile   {"",prompt="Internal use only."}                               # OLDP-4

begin

    char    l_image = "" 
	char    l_output = ""
	char    l_rawpath = ""          
	char    l_sci_ext = ""
	char    l_observatory = ""
	char    l_prefix = ""
	char    l_key_detsec = ""
    char  	l_key_datasec = ""
    char    l_key_ccdsum = ""
    char    l_gaindb = ""
    bool    l_verbose, l_fl_paste, l_fl_bias, l_fl_sat, l_fl_imexam
    int     l_gap, l_frame
    real    l_z1, l_z2, l_signal

    char    tmpin, l_dsec, l_datasec, l_stdimage, junk, tmpout, l_paste
    char    tmpim, l_imext, tmpdisp, keyfound, gaindbname
    char    tmpsignal, tmpbpm, tmpxbpm, tmpxsig
    char    l_pixtype
    int     n_ext, Xmin, Xmax, Ymin, Ymax, x1, x2, y1, y2, n_i, n_j, n_k
    int     Xoff, Yoff, Xbin, Ybin
    real    l_xcenter, l_ycenter, l_mag
    bool    fl_erase, fl_zscale, fl_zrange, l_fl_mag, l_fl_chmag, l_fl_sec
    struct  l_struct
    real    l_bias[6],bias,l_gain,l_rdnoise, l_maxbias, l_saturation
    string  l_readmode,l_gainmode,dum,s_bias, dbpm, doverlay, l_express, amp
    int     l_igain
    char    lastchar, siteprefix, utdate, imstring, cursorinput, dst	
	int     utmidnight, uthour, day,mon,yr,dymnth[12]

	# Days per month
    dymnth[1]=31 ; dymnth[2]=28 ; dymnth[3]=31 ; dymnth[4]=30 ; dymnth[5]=31
    dymnth[6]=30 ; dymnth[7]=31 ; dymnth[8]=31 ; dymnth[9]=30 ; dymnth[10]=31
    dymnth[11]=30 ; dymnth[12]=31 
	
    junk = fscan (image,       l_image)
	junk = fscan (output,      l_output)
	junk = fscan (rawpath,     l_rawpath)
    junk = fscan (observatory, l_observatory)
	junk = fscan (prefix,      l_prefix)
	junk = fscan (key_detsec,  l_key_detsec)
	junk = fscan (key_datasec, l_key_datasec)
	junk = fscan (key_ccdsum,  l_key_ccdsum)
	junk = fscan (gaindb,      l_gaindb)
	junk = fscan (sci_ext,     l_sci_ext)
    
    print (l_sci_ext) | translit ("STDIN", "a-z", "A-Z", delete-, collapse-) |\
        scan (l_sci_ext)
	
	status = 0
    l_z1 = z1
    l_z2 = z2
    l_verbose = verbose
    l_gap = gap
    l_frame = frame
    l_fl_paste = fl_paste
    Xmin = 1
    Xmax = 6144 + (2. * l_gap)
    Ymin = 1
    Ymax = 4608
    l_fl_bias = fl_bias
    l_fl_sat = fl_sat
	l_fl_imexam = fl_imexam
    l_signal = signal
        
    l_saturation = 45000.   # left hardwired since in the future it should 
                            # come from a database

    #tmpin="uparm$"//mktemp("tmpin")
    tmpin = mktemp ("tmpin")
    tmpout = mktemp ("tmpout")
    tmpdisp = mktemp ("tmpdisp")
    tmpim = mktemp ("tmpim1")
    tmpbpm = mktemp ("tmpbpm")
    tmpxbpm = mktemp ("tmpxbpm")
    tmpsignal = mktemp ("tmpsignal")
    tmpxsig = mktemp ("tmpxsig")

    fl_zscale = yes
    fl_zrange = yes
    if ((l_z1 != 0.) || (l_z2 != 0.)) {
        fl_zscale = no
        fl_zrange = no
    }

    cache ("imgets", "gimverify", "fparse", "gemisnumber")
    
	if (l_verbose) print ("...checking that the paths have a trailing slash...")
    lastchar = substr(l_rawpath, strlen(l_rawpath), strlen(l_rawpath))
    if ((l_rawpath != "") && (lastchar != "/") && (lastchar != "$"))
        l_rawpath = l_rawpath//"/"
    
	date | scan (l_struct)
	#-----------------------------------------------------------------------
    # Build file prefix for future use
    if (l_prefix == "auto") {
    
    l_observatory = strlwr (l_observatory)   # convert to lower case

    # Determine site prefix
    if (l_observatory == "gemini-north") {
        siteprefix = "N"
        utmidnight = 0
    } else if (l_observatory == "gemini-south") {
        siteprefix = "S"
        utmidnight = 20      # artificial ut midnight for file names
    } else {
        print ("ERROR:  unknown observatory: "// l_observatory)
        goto crash
    }

    gemdate (zone="UT")
    day = int(substr(gemdate.outdate,9,10))
    if (l_verbose) print ("...day = ",day)
    mon = int(substr(gemdate.outdate,6,7))
    if (l_verbose) print ("...mon = ",mon)
    yr = int(substr(gemdate.outdate,1,4))
    if (l_verbose) print ("...yr = ",yr)

    # leap year?
    if (mod(yr,4) == 0 && (mod(yr,100) != 0 || mod(yr,400) == 0)) {
        dymnth[2]=29
    }

    if (utmidnight != 0) {
        uthour = int(substr(gemdate.outdate,12,13))
        if (uthour >= utmidnight)
            day = day + 1
        if (day > dymnth[mon]) {
            mon=mon+1
            day=1
        }
        if (mon > 12) {
            yr=yr+1
            mon=1
        }
    }
    printf ("%04d%02d%02d\n",yr, mon, day) | scan (utdate)

    l_prefix = siteprefix//utdate//"S"
    }

#-----------------------------------------------------------------------
# Construct image name

    gemisnumber (l_image, "integer", verbose-)
    if (gemisnumber.fl_istype) {
        print ("GDISPLAY: Constructing image name based on today's UT \
				date or given file prefix.")

        if (l_verbose) print ("...prefix = ",l_prefix)
        print ("l_image = ", l_image)
		
        printf ("%04d\n", l_image) | scan(imstring)
        l_image = l_prefix//imstring//".fits"
        if (l_verbose) print ("...image = ",l_image)

    }

    # Is there a directory name
    gimverify (l_image)
    fparse (l_image)
    if (fparse.directory == "") {
        gimverify (l_image)                          # current directory
        if (gimverify.status != 0)
            gimverify (l_rawpath//l_image)            # try rawpath
    }
    if (gimverify.status == 1) {
        print ("ERROR: Cannot access image "//l_image)
        goto crash
    } else if (gimverify.status != 0) {
        print ("ERROR: Image "//l_image//" is not a MEF file")
        goto crash
    }
    l_image = gimverify.outname                # make sure .fits is gone

    print ("GDISPLAY: Using image "//l_image)

    fparse (gimverify.outname)
    l_imext = fparse.root
    tmpout = tmpout//l_imext
    fparse (l_image)
    if (fparse.extension != ".fits")
        l_image = l_image//".fits"
    if (fl_paste) 
        print ("GDISPLAY: Temporary image name "//tmpout)

    show stdimage | scan (l_stdimage)
    printf ("stdimage = %s\n", l_stdimage)
    if ((substr(l_stdimage,1,7) == "imtgmos") && (l_stdimage != "imtgmosccd")) {
        #    l_imX=6400 ; l_imY=4644
        Xoff = nint(6400. - Xmax) / 2.
        Yoff = nint(4644. - Ymax) / 2.
        Xmax = 6400
        Ymax = 4644
    } else {
        print ("ERROR - GDISPLAY: Please set stdimage to a full display GMOS \
            device")
        print ("                  imtgmos, imtgmos2, imtgmos4 or imtgmos8")
        goto crash
    }
    l_mag = 1.
    if (strlen(l_stdimage) == 8) 
        l_mag = 1./real(substr(l_stdimage,8,8))

    if (l_fl_bias || (l_signal != INDEF)) {
        # Select and check gaindb file
        if (l_gaindb == "default") {
            ggdbhelper (l_image//"[0]")
            if (ggdbhelper.status != 0) goto crash
            gaindbname= osfn(ggdbhelper.gaindb)            
        } else {
            gaindbname = osfn(l_gaindb)
        }
        if (access(gaindbname) == no) {
            printlog ("ERROR - GGAIN: Gain database not found", l_logfile, yes)
            goto crash
        }
    }
    
    # Check if the image has been cut.  Cannot bias if image has been cut.
    if (l_fl_bias) {
        keyfound = ""
        hselect (l_image//"[0]", "GSCUT", yes) | scan (keyfound)
        if (keyfound != "") {	#keyword was found
            print ("WARNING - GDISPLAY: Cannot bias subtract since image has \
                already been 'GSCUT'")
            print ("WARNING - GDISPLAY: Reseting fl_bias to 'no'")
            l_fl_bias = no
        }
    }

    if (!l_fl_paste && l_fl_bias) 
        print ("WARNING - GDISPLAY: Cannot bias subtract if not pasting")

    if (l_sci_ext != "ALL") {
        fxhead (l_image) | match ("IMAGE", "STDIN", stop-) | \
            translit ("STDIN", "a-z", "A-Z", delete-, collapse-) | \
            match (l_sci_ext, "STDIN", stop-, > tmpin)
    } else {
        n_k = 0
        fxhead (l_image) | match("IMAGE", "STDIN", stop-) | \
            translit ("STDIN", "a-z", "A-Z", delete-, collapse-) | \
            match ("SCI", "STDIN", stop-) | scan (n_k)
        if (n_k > 0) {
            fxhead (l_image) | match ("IMAGE", "STDIN", stop-) | \
                translit ("STDIN", "a-z", "A-Z", delete-, collapse-) | \
                match ("SCI", "STDIN", stop-, > tmpin)
            print ("Displaying the SCI extensions")
            l_sci_ext = "SCI"
        } else {
            fxhead (l_image) | match ("IMAGE", "STDIN", stop-, > tmpin)
            # raw data
            l_sci_ext = ""
        }
    }

    count (tmpin) | scan (n_ext)  # number of extensions
	
	if (n_ext == 1  && l_fl_paste) {
        print ("WARNING - GDISPLAY: only one science extension found, \
        turning paste off")
        l_fl_paste = no
    }        
    # Initialize
    l_fl_chmag = yes  # check magnification on first round
    l_fl_mag = no     # initialize check for changed magnification
    for (n_k = 1; n_k <= 6; n_k = n_k+1) {
        l_bias[n_k] = 0.0
    }


    s_bias = ""
    n_k = 0
    l_fl_sec = yes
    scanfile = tmpin
    while (fscan(scanfile, n_i) != EOF) {
        n_k = n_k + 1

        # Get binning
        imgets (l_image//"["//str(n_i)//"]", l_key_ccdsum)
        if (imgets.value == "0") {
            Xbin = 1
            Ybin = 1
        } else {
            print (imgets.value) | \
                fields ("STDIN", "1,2", lines="1", quit_if_miss-, 
                print_file-) | scan (Xbin, Ybin)
        }

        l_dsec = "0"
        l_datasec = "0"
        hselect (l_image//"["//str(n_i)//"]", l_key_detsec, yes) | scan (l_dsec)
        hselect (l_image//"["//str(n_i)//"]", l_key_datasec, yes) | scan (l_datasec)
        if ((l_dsec == "0") || (l_datasec == "0"))  {
            l_fl_sec = no
            l_dsec = ""
            l_datasec = ""
        }

        # Get the readout mode  (slow/fast lowgain modes only)
        # also needed if flagging based on signal
        if (l_fl_bias || (l_signal != INDEF)) {
            if (n_k == 1) {
                imgets (l_image//"[0]", "AMPINTEG", >& "dev$null")
                if (imgets.value == "1000")
                    l_readmode = "fast"
                else
                    l_readmode = "slow"

                # Set gainmode (high or low)  
                # If GAINORG is there use it, otherwise use GAIN
                keypar (l_image//"["//str(n_i)//"]", "GAINORIG", silent+)
                if (!keypar.found) 
                    imgets (l_image//"["//str(n_i)//"]", "GAIN", >& "dev$null")
                else
                    imgets (l_image//"["//str(n_i)//"]", "GAINORIG",
                        >& "dev$null")	 

                if (real(imgets.value) > 3.)
                    l_gainmode = "high"
                else
                    l_gainmode="low"

            }

            imgets (l_image//"["//n_i//"]", "AMPNAME", >& "dev$null" )
            l_gain = 0.0
            l_rdnoise = 0.0
            match (l_readmode, gaindbname, stop-) | \
                match (l_gainmode, "STDIN", stop-) | \
                match (imgets.value, "STDIN", stop-) | \
                scan (dum, dum, l_gain, l_rdnoise, bias, amp)
            l_bias[n_k] = bias
            if (n_k > 1)
                l_maxbias = max (l_bias[n_k], l_bias[n_k-1])
            else
                l_maxbias = l_bias[n_k]
        }
	
        if ((l_fl_paste || l_fl_bias) && !l_fl_sec) {
            print ("WARNING - GDISPLAY: No data section info found, or image \
                is already mosaiced")
            print ("                    Displaying input image. Using \
                fl_paste=no fl_bias=no")
            l_fl_paste = no
            l_fl_bias = no
        }

        # Make overlay images as needed
        if (l_fl_sat) {
            if (n_k == 1) {
                tmpxbpm = mktemp ("tmpxbpm")
                copy (l_image, tmpxbpm//".fits", verbose-)
            }
            l_express = "a>="//str(l_saturation)//" ? 1 : 0"
            imexpr (l_express, tmpxbpm//"["//str(n_i)//",overwrite]",
                l_image//"["//str(n_i)//"]", verbose-)
            hedit (tmpxbpm//"["//str(n_i)//"]", "EXTNAME", "DQ", add+,
                addonly-, delete-, verify-, show-, update+)
            hedit (tmpxbpm//"["//str(n_i)//"]", "EXTVER", str(n_k), add+,
                addonly-, delete-, verify-, show-, update+)
        }
        if (l_signal != INDEF) {
            if (n_k == 1) {
                tmpxsig = mktemp ("tmpxsig")
                copy (l_image, tmpxsig//".fits", verbose-)
            }
            l_express = "a>="//str(l_signal+l_bias[n_k])//" ? 1 : 0"
            imexpr (l_express, tmpxsig//"["//str(n_i)//",overwrite]",
                l_image//"["//str(n_i)//"]", verbose-)
            hedit (tmpxsig//"["//str(n_i)//"]", "EXTNAME", "VAR", add+,
                addonly-, delete-, verify-, show-, update+)
            hedit (tmpxsig//"["//str(n_i)//"]", "EXTVER", str(n_k), add+,
                addonly-, delete-, verify-, show-, update+)
        }
        dbpm = ""
        doverlay = ""

        if (!l_fl_paste) {

            if (l_fl_sec) {
                # find x1, x2, y1, y2 and set display parameters
                n_j = stridx (":", l_dsec)
                x1 = int(substr (l_dsec, 2, n_j-1))
                l_dsec = substr (l_dsec, n_j+1, strlen(l_dsec))
                n_j = stridx (",", l_dsec)
                x2 = int(substr (l_dsec, 1, n_j-1))
                l_dsec = substr (l_dsec, n_j+1, strlen(l_dsec))
                n_j = stridx (":", l_dsec)
                y1 = int(substr (l_dsec, 1, n_j-1))
                l_dsec = substr (l_dsec, n_j+1, strlen(l_dsec))
                y2 = int(substr (l_dsec, 1, strlen(l_dsec)-1))
            } else {
                x1 = 1
                y1 = 1
                hselect (l_image//"["//str(n_i)//"]", "i_naxis1,i_naxis2",
                    yes) | scan (x2, y2)
                x2 = x2 * Xbin
                y2 = y2 * Ybin
            }

            if (n_k == 1)
                fl_erase = yes
            else
                fl_erase = no

            l_xcenter = (x2-x1)/2. + x1
            if (l_xcenter > 2048)
                l_xcenter += l_gap
            if (l_xcenter > (2*2048. + l_gap))
                l_xcenter += l_gap
            l_xcenter += Xoff

            l_xcenter = l_xcenter / Xmax
            l_ycenter = ((y2-y1)/2. + y1 + Yoff) / Ymax

            # Reset xcenter,ycenter if image is binned
            l_xcenter = (l_xcenter-0.5)/Xbin + 0.5
            l_ycenter = (l_ycenter-0.5)/Ybin + 0.5

            # Reset magnification if higher resolution can be displayed
            # Never display more than 1pix to 1pix
            if ((min(Xbin,Ybin) > 1.) && (l_mag*min(Xbin,Ybin) <= 1.) && \
                l_fl_chmag) {
                
                l_mag = l_mag * min(Xbin,Ybin)
                l_fl_chmag = no  # don't check/change l_mag in the next rounds
                l_fl_mag = yes   # update xcenter,ycenter in all rounds
            }
            # Adjust xcenter,ycenter if l_mag has been changed
            if (l_fl_mag) {
                l_xcenter = (l_xcenter-0.5) * min(Xbin,Ybin) + 0.5
                l_ycenter = (l_ycenter-0.5) * min(Xbin,Ybin) + 0.5
            } 

            if (l_fl_sat)
                dbpm = tmpxbpm//"["//str(n_i)//"]"//l_datasec
            if (l_signal != INDEF)
                doverlay = tmpxsig//"["//str(n_i)//"]"//l_datasec
            if (l_fl_sat && (n_k == 1))
                print ("GDISPLAY: Saturated pixels flagged as red")
            if ((l_signal != INDEF) && (n_k == 1))
                print ("GDISPLAY: Pixels with signal>="//str(l_signal)//\
                    " above bias flagged as green")
            printf ("%s %4d %4d %4d %4d  %5.1f  ", l_image//"["//str(n_i)//"]",
                x1, x2, y1, y2, l_mag)

            display (l_image//"["//str(n_i)//"]"//l_datasec, l_frame,
                erase=fl_erase, xcenter=l_xcenter, ycenter=l_ycenter, z1=l_z1,
                z2=l_z2, zrange=fl_zrange, zscale=fl_zscale, xmag=l_mag,
                ymag=l_mag, fill=no, bpm=dbpm, bpdisplay="overlay",
                bpcolors="red", overlay=doverlay, ocolors="green")
		    
        } else if (l_fl_sec) {
            # Setup parameters for gmosaic, get EXTNAME,EXTVER in header
            if (n_k == 1) {
                copy (l_image, tmpim//".fits", verbose-)
                l_paste = tmpim
            }
            if (l_fl_bias) {
                imarith (l_paste//"["//str(n_i)//"]", "-", l_bias[n_k],
                    l_paste//"["//str(n_i)//",overwrite+]")
            }
            if (l_sci_ext == "") {
                hedit (l_paste//"["//str(n_i)//"]", "EXTNAME", "SCI", add+,
                    addonly-, delete-, verify-, show-, update+)
                hedit (l_paste//"["//str(n_i)//"]", "EXTVER", str(n_k), add+,
                    addonly-, delete-, verify-, show-, update+)
            }

        }

    } # end of while loop
    scanfile = ""
    flpr  # make sure we can use EXTNAME,EXTVER for access
    
    if (l_sci_ext == "")
        l_sci_ext = "SCI"

    delete (tmpin, verify-, >& "dev$null")
    
    if (n_ext == 1 && l_fl_imexam) {
        print ("GDISPLAY: Starting imexamine - quit with q")
        imexam ()
    } 
    
    if (l_fl_paste) {
        print ("GDISPLAY: Pasting images")
        gmosaic (l_paste, outimages=tmpout, fl_paste+, gap=l_gap, fl_fixpix-,
            verbose=l_verbose, fl_clean-, sci_ext=l_sci_ext, fl_real-, 
            logfile="dev$null")
         
        if (l_fl_sat) {
            tmpbpm = mktemp("tmpbpm")
            print ("GDISPLAY: Saturated pixels flagged as red")
            gmosaic (tmpxbpm, outimages=tmpbpm, fl_paste+, gap=l_gap,
                fl_fixpix-, geointer="nearest", verbose=l_verbose, fl_clean-,
                sci_ext="DQ", fl_real-, logfile="dev$null")
            dbpm = tmpbpm//"[DQ,1]"
        } else
            dbpm = ""

        if (l_signal != INDEF) {
            tmpsignal = mktemp ("tmpsignal")
            print ("GDISPLAY: Pixels with signal>="//str(l_signal)//\
                " above bias flagged as green")
            gmosaic (tmpxsig, outimages=tmpsignal, fl_paste+, gap=l_gap,
                fl_fixpix-, geointer="nearest", verbose=l_verbose, fl_clean-,
                sci_ext="VAR", fl_real-, logfile="dev$null")
            doverlay = tmpsignal//"[VAR,1]"
        } else
            doverlay = ""
        flpr # to get access through VAR,DQ after hedit

        display (tmpout//"[1]", l_frame, erase=yes, xcenter=0.5, ycenter=0.5,
            xmag=1, ymag=1, z1=l_z1, z2=l_z2, zrange=fl_zrange,
            zscale=fl_zscale, fill=no, bpm=dbpm, bpdisplay="overlay",
            bpcolors="red", overlay=doverlay, ocolors="green")
        if (l_fl_imexam) {
            print ("GDISPLAY: Starting imexamine - quit with q")
            imexamine ()
        }
        if (l_output == "")
            imdelete (tmpout, verify-, >& "dev$null")
        else
            imrename (tmpout, l_output, verbose+)

    }
    
    goto clean

crash:
    # Exit with error subroutine
    status = 1
clean:
    imdelete (tmpim//","//tmpsignal//","//tmpbpm//","//tmpxbpm//","//tmpxsig,
        ver-, >& "dev$null")
    
end
